/**
 * Copyright (c) 2021 Dominic Masters
 * 
 * This software is released under the MIT License.
 * https://opensource.org/licenses/MIT
 */

#include "holdemrender.h"

void holdemRenderInit() {
  int32_t lWidth, rWidth, height;
  card_t card;
  tilesetdiv_t *cardBack;

  // Prepare the two frame buffers.
  lWidth = HOLDEM_GAME_FRAME_LEFT_WIDTH, rWidth = HOLDEM_GAME_FRAME_RIGHT_WIDTH;
  height = HOLDEM_GAME_FRAME_HEIGHT;
  HOLDEM_GAME_STATE.frameLeft = frameBufferCreate(lWidth, height);
  HOLDEM_GAME_STATE.frameRight = frameBufferCreate(rWidth, height);
  HOLDEM_GAME_STATE.quadLeft = quadCreate(0, 0, 0, 0, 0, lWidth, height, 1, 1);
  HOLDEM_GAME_STATE.quadRight = quadCreate(0, 0, 0, 0, 0, rWidth, height, 1, 1);

  // Font
  HOLDEM_GAME_STATE.fontTexture = assetTextureLoad("font.png");
  HOLDEM_GAME_STATE.fontTileset = tilesetCreate(20, 20,
    HOLDEM_GAME_STATE.fontTexture->width,
    HOLDEM_GAME_STATE.fontTexture->height,
    1, 1, 1, 1
  );
  HOLDEM_GAME_STATE.fontBatch = spriteBatchCreate(1024);

  // Poker Table
  HOLDEM_GAME_STATE.tablePrimitive = pokerTableCreate();
  HOLDEM_GAME_STATE.tableTexture = assetTextureLoad("pokertable.png");

  // Kagami
  HOLDEM_GAME_STATE.kagamiTexture = assetTextureLoad("kagami.png");
  HOLDEM_GAME_STATE.kagamiTileset = tilesetCreate(3, 2,
    HOLDEM_GAME_STATE.kagamiTexture->width,
    HOLDEM_GAME_STATE.kagamiTexture->height,
    0, 0, 0, 0
  );
  HOLDEM_GAME_STATE.kagamiQuad = quadCreate(0, 0, 0, 0, 0, 1, 1, 1, 1);

  // Load Cards Texture
  HOLDEM_GAME_STATE.cardTexture = assetTextureLoad("cards_normal.png");
  HOLDEM_GAME_STATE.cardTileset = tilesetCreate(CARD_COUNT_PER_SUIT, 6,
    HOLDEM_GAME_STATE.cardTexture->width, HOLDEM_GAME_STATE.cardTexture->height,
    0, 0, 0, 0
  );

  // Cards Primitive
  cardBack = HOLDEM_GAME_STATE.cardTileset->divisions+(
    HOLDEM_GAME_STATE.cardTileset->columns * 4
  );
  HOLDEM_GAME_STATE.cardPrimitive = primitiveCreate(
    QUAD_VERTICE_COUNT * 2, QUAD_INDICE_COUNT * 2
  );
  quadBuffer(HOLDEM_GAME_STATE.cardPrimitive, -HOLDEM_GAME_CARD_DEPTH,
    -HOLDEM_GAME_CARD_WIDTH, -HOLDEM_GAME_CARD_HEIGHT,
    cardBack->x0, cardBack->y1,
    HOLDEM_GAME_CARD_WIDTH, HOLDEM_GAME_CARD_HEIGHT,
    cardBack->x1, cardBack->y0,
    QUAD_VERTICE_COUNT, QUAD_INDICE_COUNT
  );
}

void holdemRender() {
  int32_t lWidth, rWidth, height;

  // Resize Frame buffers.
  lWidth = HOLDEM_GAME_FRAME_LEFT_WIDTH, rWidth = HOLDEM_GAME_FRAME_RIGHT_WIDTH;
  height = HOLDEM_GAME_FRAME_HEIGHT;
  if(
    HOLDEM_GAME_STATE.frameLeft->texture->width != lWidth ||
    HOLDEM_GAME_STATE.frameLeft->texture->height != height
  ) {
    frameBufferDispose(HOLDEM_GAME_STATE.frameLeft);
    frameBufferDispose(HOLDEM_GAME_STATE.frameRight);
    HOLDEM_GAME_STATE.frameLeft = frameBufferCreate(lWidth, height);
    HOLDEM_GAME_STATE.frameRight = frameBufferCreate(rWidth, height);
    quadBuffer(HOLDEM_GAME_STATE.quadLeft, 0,
      0, 0, 0, 1,
      lWidth, height, 1, 0,
      0, 0
    );
    quadBuffer(HOLDEM_GAME_STATE.quadRight, 0,
      0, 0, 0, 1,
      rWidth, height, 1, 0,
      0, 0
    );
  }

  // Render things on the left frame buffer
  glClearColor(0.3, 0, 0, 1);
  frameBufferUse(HOLDEM_GAME_STATE.frameLeft, true);
  cameraPerspective(&HOLDEM_GAME_STATE.cameraLeft, 25, 
    ((float)lWidth/height), 0.2f, 1000.0f
  );
  // cameraLookAt(&HOLDEM_GAME_STATE.cameraLeft, 2, 2, 2, 0, 0, 0);
  holdemLookHand(&HOLDEM_GAME_STATE.cameraLeft, 0x00);
  shaderUseCamera(GAME_STATE.shaderWorld, &HOLDEM_GAME_STATE.cameraLeft);
  holdemRenderWorld();

  // Render things on the right frame buffer
  glClearColor(0.3, 0.3, 0, 1);
  frameBufferUse(HOLDEM_GAME_STATE.frameRight, true);
  cameraPerspective(&HOLDEM_GAME_STATE.cameraRight, 45, 
    ((float)rWidth/height), 0.25f, 100.0f
  );
  cameraLookAt(&HOLDEM_GAME_STATE.cameraRight, 0, 3, 3, 0, 0, 0);
  shaderUseCamera(GAME_STATE.shaderWorld, &HOLDEM_GAME_STATE.cameraRight);
  holdemRenderWorld();

  // Finally, render the frame buffers to the back buffer.
  glClearColor(0, 0, 0, 1);
  frameBufferUse(NULL, true);
  cameraOrtho(&GAME_STATE.cameraWorld, 0,
    RENDER_STATE.width, RENDER_STATE.height, 1, 0, 1
  );
  cameraLookAt(&GAME_STATE.cameraWorld, 0, 0, 0.5f, 0, 0, 0);
  shaderUseCamera(GAME_STATE.shaderWorld, &GAME_STATE.cameraWorld);
  shaderUsePosition(GAME_STATE.shaderWorld,
    0, 0, 0, 0, 0, 0
  );
  shaderUseTexture(GAME_STATE.shaderWorld, HOLDEM_GAME_STATE.frameLeft->texture);
  primitiveDraw(HOLDEM_GAME_STATE.quadLeft, 0, -1);
  shaderUsePosition(GAME_STATE.shaderWorld,
    RENDER_STATE.width-rWidth, 0, 0, 0, 0, 0
  );
  shaderUseTexture(GAME_STATE.shaderWorld, HOLDEM_GAME_STATE.frameRight->texture);
  primitiveDraw(HOLDEM_GAME_STATE.quadRight, 0, -1);
}

void holdemLookSeat(camera_t *camera, uint8_t seat) {
  float x, z, angle;

  angle = mathDeg2Rad(-45*seat);
  x = sin(angle);
  z = cos(angle);
  cameraLookAt(&HOLDEM_GAME_STATE.cameraLeft,
     x, 0.2,  z,
    -x, 0.2, -z
  );
}

void holdemLookHand(camera_t *camera, uint8_t seat) {
  float x, lx, z, lz, angle, distance;

  angle = mathDeg2Rad(-45*seat);
  x = sin(angle);
  z = cos(angle);

  cameraLookAt(&HOLDEM_GAME_STATE.cameraLeft,
    x*0.1, 0.8, z*0.1,
    -x*0.5, 0.2, -z*0.5
  );
}

void holdemRenderPlayer(float x, float y, float z, float yaw) {
  float w, h;
  w = 0.6, h = (
    (float)HOLDEM_GAME_STATE.kagamiTileset->divY /
    (float)HOLDEM_GAME_STATE.kagamiTileset->divX
  ) * w;
  
  int i = (int32_t)(TIME_STATE.current*10)%HOLDEM_GAME_STATE.kagamiTileset->count;
  quadBuffer(HOLDEM_GAME_STATE.kagamiQuad, 0,
    -w/2, -h/2,
    HOLDEM_GAME_STATE.kagamiTileset->divisions[i].x0,
    HOLDEM_GAME_STATE.kagamiTileset->divisions[i].y1,
    w/2, h/2,
    HOLDEM_GAME_STATE.kagamiTileset->divisions[i].x1,
    HOLDEM_GAME_STATE.kagamiTileset->divisions[i].y0,
    0, 0
  );

  shaderUseTexture(GAME_STATE.shaderWorld, HOLDEM_GAME_STATE.kagamiTexture);
  shaderUsePosition(GAME_STATE.shaderWorld, x, y, z, 0, mathDeg2Rad(yaw), 0);
  primitiveDraw(HOLDEM_GAME_STATE.kagamiQuad, 0, -1);
}

void holdemRenderCard(card_t card,
  float x, float y, float z, float pitch, float yaw, float roll
) {
  tilesetdiv_t *cardFront = HOLDEM_GAME_STATE.cardTileset->divisions + card;
  quadBuffer(HOLDEM_GAME_STATE.cardPrimitive, HOLDEM_GAME_CARD_DEPTH,
    -HOLDEM_GAME_CARD_WIDTH, -HOLDEM_GAME_CARD_HEIGHT, 
    cardFront->x0, cardFront->y1,
    HOLDEM_GAME_CARD_WIDTH, HOLDEM_GAME_CARD_HEIGHT,
    cardFront->x1, cardFront->y0,
    0, 0
  );

  shaderUseTexture(GAME_STATE.shaderWorld, HOLDEM_GAME_STATE.cardTexture);
  shaderUsePosition(GAME_STATE.shaderWorld, x, y, z, pitch, yaw, roll);
  primitiveDraw(HOLDEM_GAME_STATE.cardPrimitive, 0, -1);
}

void holdemRenderWorld() {
  uint8_t i, j;
  float pitch;
  holdemplayer_t *player;
  char name[16];

  // Poker Table
  shaderUsePositionAndScale(GAME_STATE.shaderWorld,
    0, -0.01, 0,
    0, 0, 0,
    3.4, 3.4, 3.4
  );
  shaderUseTexture(GAME_STATE.shaderWorld, HOLDEM_GAME_STATE.tableTexture);
  primitiveDraw(HOLDEM_GAME_STATE.tablePrimitive, 0, -1);
  
  pitch = mathDeg2Rad(-90);
  for(j = 0; j < HOLDEM_GAME_STATE.match.cardsFacing; j++) {
    holdemRenderCard(HOLDEM_GAME_STATE.match.cards[j], j*0.2, 0, -0.2, pitch, 0, 0);
  }

  for(i = 0; i < HOLDEM_PLAYER_COUNT; i++) {
    player = HOLDEM_GAME_STATE.match.players + i;
    if(player->state & HOLDEM_STATE_FOLDED) continue;
    sprintf(name, "Player %i", i);

    shaderUseTexture(GAME_STATE.shaderWorld, HOLDEM_GAME_STATE.fontTexture);
    spriteBatchFlush(HOLDEM_GAME_STATE.fontBatch);
    shaderUsePosition(GAME_STATE.shaderWorld, -0.1,0,i*0.2, mathDeg2Rad(-90),0,0);
    fontSpriteBatchBuffer(
      HOLDEM_GAME_STATE.fontBatch,
      HOLDEM_GAME_STATE.fontTileset,
      name, FONT_RIGHT_X, FONT_CENTER_Y, 0, -1, 0.1
    );
    spriteBatchDraw(HOLDEM_GAME_STATE.fontBatch, 0, -1);

    for(j = 0; j < HOLDEM_PLAYER_HAND; j++) {
      // pitch = mathDeg2Rad(player->state & HOLDEM_STATE_SHOWING ? -90 : 90);
      holdemRenderCard(player->cards[j], j*0.2, 0, i*0.2, pitch, 0, 0);
    }
  }

  shaderUseTexture(GAME_STATE.shaderWorld, HOLDEM_GAME_STATE.fontTexture);

  spriteBatchFlush(HOLDEM_GAME_STATE.fontBatch);
  shaderUsePosition(GAME_STATE.shaderWorld, -0.1,0,-0.2, mathDeg2Rad(-90),0,0);
  fontSpriteBatchBuffer(
    HOLDEM_GAME_STATE.fontBatch,
    HOLDEM_GAME_STATE.fontTileset,
    "Dealer", FONT_RIGHT_X, FONT_CENTER_Y, 0, -1, 0.1
  );
  spriteBatchDraw(HOLDEM_GAME_STATE.fontBatch, 0, -1);

  // Draw the players
  float playerY = 0.34;
  holdemRenderPlayer(0, playerY, -1, 0);
  // holdemRenderPlayer(-1, playerY, -0, 90);
  // holdemRenderPlayer(-0.75, playerY, 0.75, -45);
  // holdemRenderPlayer(0.75, playerY, 0.75, 45);
  // holdemRenderPlayer(1, playerY, 0, -90);
}
